@@ -13,6 +13,7 @@
 # OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
 import csv
+import sys
 import os
 import requests
 import tempfile
@@ -27,16 +28,19 @@ from cartman import ui
 from cartman import text
 
 
+BASE_DIRECTORY = "~/.cartman"
 CONFIG_LOCATIONS = [
-    os.path.expanduser("~/.cartman/config"),
+    os.path.expanduser(os.path.join(BASE_DIRECTORY, "config")),
     "/etc/cartmanrc",
 ]
 
 MIN_TRAC_VERSION = (0, 11)
 MAX_TRAC_VERSION = (1, 0)
 
-AUTH_TYPES = {'basic': requests.auth.HTTPBasicAuth,
-              'digest': requests.auth.HTTPDigestAuth}
+AUTH_TYPES = {
+    "basic": requests.auth.HTTPBasicAuth,
+    "digest": requests.auth.HTTPDigestAuth,
+}
 
 DEFAULT_TEMPLATE = """To:
 Cc:
@@ -62,17 +66,85 @@ class CartmanApp(object):
         self.site = "trac"
         self.logged_in = False
         self.browser = webbrowser
+        self.trac_version = (0, 0)
 
-    def _read_config(self):
+    def run(self, args):
+        """Main function call.
+
+        Converts the options and arguments into a function call within this
+        instance.
+
+        :param options: Options returned from the optparse module.
+        :param args: Arguments returned from the optparse module.
+
+        """
+        self.site = args.site or self.site
+        self.open_after = args.open_after
+        self.add_comment = args.add_comment
+        self.message = args.message
+        self.template = args.template
+
+        self.ensure_directories()
+        self.read_config()
+        self.session = requests.session()
+
+        auth_class = AUTH_TYPES[self.auth_type]
+        self.session.auth = auth_class(self.username, self.password)
+        self.session.verify = self.verify_ssl_cert
+
+        func_name = "run_" + args.command
+        func = getattr(self, "run_" + args.command, None)
+        if not func:
+            raise exceptions.UnknownCommand("unknown command: " + args.command)
+
+        if "help" in args.parameters:
+            self.print_function_help(func_name)
+            return
+
+        self.login()
+
+        try:
+            output = func(*args.parameters)
+        except exceptions.InvalidParameter as ex:
+            print("error: {}\n".format(ex))
+            self.print_function_help(func_name)
+            return
+
+        self.print_output(output)
+
+    def print_output(self, output):
+        if output:
+            print("\n".join(output))
+
+    def ensure_directories(self):
+        """Creates a ~/.cartman/ if none exist."""
+
+        expanded_directory = os.path.expanduser(BASE_DIRECTORY)
+
+        if os.path.exists(expanded_directory):
+            return
+
+        os.mkdir(expanded_directory, 0o750)
+
+    def read_config(self):
+        """Populate the instance with settings for the config file.
+
+        If we can't find any section for the given site, error gracefully.
+
+        """
         defaults = {
             "auth_type": "basic",
-            "verify_ssl_cert": "true"
+            "verify_ssl_cert": "true",
         }
 
         cp = configparser.RawConfigParser(defaults)
         cp.read(CONFIG_LOCATIONS)
 
-        self.base_url = cp.get(self.site, "base_url")
+        if not cp.has_option(self.site, "base_url"):
+            raise exceptions.ConfigError("unable to find a [{}] section with "
+                                         "a base_url.".format(self.site))
+
+        self.base_url = cp.get(self.site, "base_url").rstrip("/")
         self.username = cp.get(self.site, "username")
         self.password = cp.get(self.site, "password")
         self.verify_ssl_cert = cp.getboolean(self.site, "verify_ssl_cert")
@@ -81,16 +153,18 @@ class CartmanApp(object):
         auth_type = cp.get(self.site, "auth_type")
         auth_type = auth_type.lower()
         if auth_type not in AUTH_TYPES:
-            msg = ("Invalid auth setting: {}. Allowed auth_type values are: {}"
-                   .format(auth_type, ', '.join(sorted(AUTH_TYPES.keys()))))
-            raise exceptions.InvalidConfigSetting(msg)
+            supported_auths = ", ".join(sorted(AUTH_TYPES.keys()))
+            msg = ("invalid auth setting '{}', supported: {}"
+                  .format(auth_type, supported_auths))
+            raise exceptions.ConfigError(msg)
         self.auth_type = auth_type
 
         self.required_fields = ["To", "Component", "Subject", "Priority"]
 
-    def _get_form_token(self):
+    def get_form_token(self):
         """Return the form_token sent on all the POST forms for validation.
-        This value is store as a cookie, on the session. If the specifically
+
+        This value is stored as a cookie, on the session. If the specifically
         named cookie is not found, returns an empty string.
 
         """
@@ -100,52 +174,102 @@ class CartmanApp(object):
 
         return ""
 
-    def _get_properties(self):
-        """Return all the values typically used in drop-downs on the create
-        ticket page, such as Milestones, Versions, etc. These lists are
-        extracted from the JavaScript dictionary exposed on the query page.
+    def get_properties(self):
+        """Return the values used in drop-downs on the create ticket page.
+
+        The lists such as Milestones and Versions are extracted from a
+        JavaScript dictionary exposed on the query page.
 
         """
         return text.extract_properties(self.get("/query").text)
 
-    def _check_version(self):
-        if self.trac_version < MIN_TRAC_VERSION \
-                or self.trac_version > MAX_TRAC_VERSION:
-            version = ".".join([str(tok) for tok in self.trac_version])
+    def check_version(self, raw_html):
+        """Print a warning if the version of Trac is unsupported."""
+
+        # If we have a version in there, we likely already figure this out.
+        if self.trac_version != (0, 0):
+            return
+
+        trac_version = text.extract_trac_version(raw_html)
+
+        # We can't extract anything from this query.
+        if not trac_version:
+            return
+
+        if trac_version < MIN_TRAC_VERSION or trac_version > MAX_TRAC_VERSION:
+            version = ".".join([str(tok) for tok in trac_version])
             print("WARNING: Untested Trac version ({})".format(version))
 
-    def _editor(self, filename):
+        self.trac_version = trac_version
+
+    def editor(self, filename):
+        """Spawn the default editor ($EDITOR env var)."""
+
+        if not os.getenv("EDITOR"):
+            raise exceptions.FatalError("unable to get an EDITOR environment "
+                                        "variable")
+
         os.system("$EDITOR '{}'".format(filename))
 
-    def _input(self, prompt):
+    def input(self, prompt):
         return raw_input(prompt)
 
-    def get(self, query_string, data=None):
+    def get(self, query_string, data=None, handle_errors=True):
         """Generates a GET query on the target Trac system.
 
+        TODO: extract all the possible error elements as message.
+
         :param query_string: Starts with a slash, part of the URL between the
                              domain and the parameters (before the ?).
         :param data: Dictionary of parameters to encode at the end of the
                      ``query_string``.
-        """
-        return self.session.get(self.base_url + query_string, data=data)
+        :param handle_errors: Crash with a proper exception according to the
+                              HTTP return code (default: True).
 
-    def post(self, query_string, data=None):
-        """Generates a POST query on the target Trac system. This also alters
-        the given data to include the form token stored on the cookies. Without
-        this token, Trac will refuse form submissions.
+        """
+        r = self.session.get(self.base_url + query_string, data=data)
+
+        if r.status_code >= 400 and handle_errors:
+            message = text.extract_message(r.text)
+            if not message:
+                message = "{} returned {}".format(self.base_url, r)
+            raise exceptions.FatalError(message)
+
+        # Check the version if we can.
+        self.check_version(r.text)
+
+        return r
+
+    def post(self, query_string, data=None, handle_errors=True):
+        """Generates a POST query on the target Trac system.
+
+        This also alters the given data to include the form token stored on the
+        cookies. Without this token, Trac will refuse form submissions.
 
         :param query_string: Starts with a slash, part of the URL between the
                              domain and the parameters (before the ?).
         :param data: Dictionary of parameters to encode and transmit to the
                      target page.
+        :param handle_errors: Crash with a proper exception according to the
+                              HTTP return code (default: True).
+
         """
         if data:
-            data["__FORM_TOKEN"] = self._get_form_token()
-        return self.session.post(self.base_url + query_string, data=data)
+            data["__FORM_TOKEN"] = self.get_form_token()
+
+        r = self.session.post(self.base_url + query_string, data=data)
+
+        if r.status_code >= 400 and handle_errors:
+            message = text.extract_message(r.text)
+            if not message:
+                message = "{} returned {}".format(self.base_url, r)
+            raise exceptions.FatalError(message)
+
+        return r
 
     def login(self):
         """Ensure the current session is logged-in, accessing the main page.
+
         This will allow Trac to generate the cookies that will be stored on our
         ``self.session`` object.
 
@@ -155,49 +279,44 @@ class CartmanApp(object):
 
         # Seems that depending on the method used to serve trac, we need to use
         # a different path to initiate authentication.
-        r = self.get("/login")
+        r = self.get("/login", handle_errors=False)
 
-        if r.status_code == 401:
-            msg = ("login failed on {}. You might be using wrong "
-                   "username/password combo or wrong auth type"
+        if r.status_code not in (200, 302):
+            msg = ("login failed on {} (bad user, password or auth type)"
                    .format(r.request.url))
             raise exceptions.LoginError(msg)
 
+        # Load a page to get the new cookies.
         if r.status_code != 302:
             r = self.get("/")
 
-        if r.status_code not in (200, 302):
-            raise exceptions.LoginError("login failed on {}"
-                                        .format(r.request.url))
-
-        # Grab the Trac version number, and throw a warning if not 0.12.
-        self.trac_version = text.extract_trac_version(r.text)
-        self._check_version()
-
         self.logged_in = True
 
     def get_dicts(self, query_string):
-        """Wrapper around the ``get`` method that ensures we are logged-in and
-        converts the returned data to dictionaries, assuming the query
-        originally returns tab-delimited raw data.
+        """Wrapper around ``get()`` that ensures auth and returns dicts.
+
+        This methods assumes the response contains tab-delimited data.
 
         :param query_string: Starts with a slash, part of the URL between the
                              domain and the parameters (before the ?).
-        """
 
-        self.login()
+        """
         r = self.get(query_string)
-        data = r.content
+        data = r.text
 
         # Recent version of Trac seem to be sending data with a BOM (?!)
-        if data[0] == u'\ufeff':
+        if data[0] == u"\ufeff":
             data = data[1:]
 
+        # And since the csv module in Python 2.7 is not unicode-friendly, we
+        # encode to UTF-8.
+        if sys.version < '3':
+            data = data.encode("utf-8")
+
         return csv.DictReader(data.splitlines(True), delimiter="\t")
 
     def get_tickets(self, query_string):
-        """Wrapper around the ``get_dicts`` method that converts the
-        dictionaries into ``Ticket`` instances.
+        """Wrapper around ``get_dicts()`` that returns ``Ticket`` instances.
 
         :param query_string: Starts with a slash, part of the URL between the
                              domain and the parameters (before the ?).
@@ -217,6 +336,7 @@ class CartmanApp(object):
 
     def list_commands(self):
         """Return list of all the commands."""
+
         commands = []
         for attrname in dir(self):
             if attrname.startswith("run_"):
@@ -225,6 +345,7 @@ class CartmanApp(object):
 
     def print_commands_list(self):
         """Print command list minus help."""
+
         print("Available Commands:\n")
         for command_name in self.list_commands():
             if command_name != "run_help":
@@ -234,6 +355,7 @@ class CartmanApp(object):
     # TODO fix indent on doc strings..
     def print_commands(self):
         """Initial attempt to return a help screen with all the commands."""
+
         for command_name in self.list_commands():
             self.print_function_help(command_name)
             print("")
@@ -291,142 +413,40 @@ class CartmanApp(object):
         """
         return "\n".join([": ".join(pair) for pair in headers.items()])
 
-    def run(self, args):
-        """Main function called, convert the options and arguments into a
-        function call within this instance.
-
-        :param options: Options returned from the optparse module.
-        :param args: Arguments returned from the optparse module.
-
-        """
-        self.site = args.site or self.site
-        self.open_after = args.open_after
-        self.add_comment = args.add_comment
-        self.message = args.message
-        self.template = args.template
-
-        self._read_config()
-        self.session = requests.session()
-
-        auth_class = AUTH_TYPES[self.auth_type]
-        self.session.auth = auth_class(self.username, self.password)
-        self.session.verify = self.verify_ssl_cert
-
-        func_name = "run_" + args.command
-        if hasattr(self, func_name):
-            func = getattr(self, func_name)
-            if "help" in args.parameters:
-                self.print_function_help(func_name)
-                return
-
-            try:
-                func(*args.parameters)
-            except exceptions.InvalidParameter as ex:
-                print("error: {}\n".format(ex))
-                self.print_function_help(func_name)
-                return
-        else:
-            raise exceptions.UnknownCommand("unknown command: " + args.command)
-
-    def run_help(self, command="help"):
-        """Show the help for a given command.
-
-        usage: cm help command
-
-        """
-        func_name = "run_" + command
-        if hasattr(self, func_name):
-            self.print_function_help(func_name)
-            if command == "help":
-                self.print_commands_list()
-        else:
-            raise exceptions.UnknownCommand("unknown command: " + func_name)
-
-    def run_report(self, report_id=None):
-        """List tickets from a given report number.
-
-        usage: cm report ticket_id
-
-        """
-        report_id = text.validate_id(report_id)
-
-        query_string = "/report/{}?format=tab".format(report_id)
-
-        for t in self.get_tickets(query_string):
-            print(t.format_title())
-
-    def run_reports(self):
-        """List reports available in the system.
-
-        usage: cm reports
-
-        """
-        for d in self.get_dicts("/report?format=tab"):
-            print("#{report}. {title}".format(**d))
-
-    def run_view(self, ticket_id):
-        """Display a ticket summary.
-
-        usage: cm view ticket_id
-
-        """
-        ticket_id = text.validate_id(ticket_id)
-
-        query_string = "/ticket/{}?format=tab".format(ticket_id)
-
-        t = next(self.get_tickets(query_string))
-        title = t.format_title()
-
-        print(ui.title(title))
-        print("")
-
-        print(t.description)
-
-    def run_open(self, ticket_id):
-        """Open a ticket in your browser.
-
-        usage: cm open ticket_id
-        """
-        ticket_id = text.validate_id(ticket_id)
-
-        self.open_in_browser(ticket_id)
-
-    def run_properties(self):
-        """Lists the system's properties (Milestone, Component, etc.).
-
-        usage: cm properties
-
-        """
-        properties = self._get_properties()
-
-        print(ui.title("Milestones"))
-        print(", ".join(properties["milestone"]["options"]) + "\n")
-
-        print(ui.title("Components"))
-        print(", ".join(properties["component"]["options"]) + "\n")
-
-        print(ui.title("Status"))
-        print(", ".join(properties["status"]["options"]) + "\n")
-
-        print(ui.title("Priority"))
-        print(", ".join(properties["priority"]["options"]) + "\n")
-
     def _read_comment(self):
         """Prompt for a piece of text via the current EDITOR. Returns a string.
         """
         (fd, filename) = tempfile.mkstemp(suffix=".cm.ticket")
-        self._editor(filename)
+        self.editor(filename)
         with open(filename) as fp:
             comment = fp.read()
         return comment
 
     def _extract_timestamps(self, raw_html):
         """Wrapper around extract_time_v0 and extract_time_v1."""
+
         if self.trac_version >= (1, 0):
             return text.extract_timestamps_v1(raw_html)
         else:
             return text.extract_timestamps_v0(raw_html)
 
+    def extract_status_from_ticket_page(self, raw_html):
+        if self.trac_version >= (1, 0):
+            return text.extract_status_from_ticket_page_v1(raw_html)
+        else:
+            return text.extract_status_from_ticket_page_v0(raw_html)
+
+    #
+    # Command definitions
+    #
+    # The following commands are triggered when cm is called from the command
+    # line. The first argument is the command, the corresponding method is
+    # named with a run_ prefix.
+    #
+    # Note that the docstring for these functions is used for the --help auto
+    # documentation.
+    #
+
     def run_comment(self, ticket_id):
         """Add a comment to the given ticket_id. This command does not return
         anything if successful. Command is cancelled if the content of the
@@ -437,8 +457,6 @@ class CartmanApp(object):
         """
         ticket_id = text.validate_id(ticket_id)
 
-        self.login()
-
         # Load the timestamps from the ticket page.
         r = self.get("/ticket/{}".format(ticket_id))
         timestamps = self._extract_timestamps(r.text)
@@ -465,57 +483,24 @@ class CartmanApp(object):
         if self.trac_version >= (1, 0):
             token = 'system-message" style=""'
         else:
-            token = 'system-message'
+            token = "system-message"
 
         if token in r.text or r.status_code != 200:
             raise exceptions.FatalError("unable to save comment")
 
-    def run_status(self, ticket_id, status=None):
-        """Updates the status of a ticket.
+    def run_help(self, command="help"):
+        """Show the help for a given command.
 
-        usage: cm status ticket_id [new_status]
+        usage: cm help command
 
         """
-        ticket_id = text.validate_id(ticket_id)
-
-        self.login()
-
-        # Get all the available actions for this ticket
-        r = self.get("/ticket/{}".format(ticket_id))
-        timestamps = self._extract_timestamps(r.text)
-        statuses = text.extract_statuses(r.text)
-
-        # A ``status`` was provided, try to find the exact match, else just
-        # display the current status for this ticket, and the available ones.
-        if status:
-            status = text.fuzzy_find(status, statuses)
-
-            if not status:
-                raise exceptions.FatalError("bad status (for this ticket: {})"
-                                            .format(", ".join(statuses)))
+        func_name = "run_" + command
+        if hasattr(self, func_name):
+            self.print_function_help(func_name)
+            if command == "help":
+                self.print_commands_list()
         else:
-            status = text.extract_status_from_ticket_page(r.text)
-            print("Current status: {}".format(status))
-            print("Available statuses: {}".format(", ".join(statuses)))
-            return
-
-        if self.message:
-            comment = self.message
-        elif self.add_comment:
-            comment = self._read_comment()
-        else:
-            comment = ""
-
-        data = {
-            "action": status,
-            "comment": comment,
-        }
-        data.update(timestamps)
-
-        r = self.post("/ticket/{}".format(ticket_id), data)
-
-        if "system-message" in r.text or r.status_code != 200:
-            raise exceptions.FatalError("unable to set status")
+            raise exceptions.UnknownCommand("unknown command: " + func_name)
 
     def run_new(self, owner=None):
         """Create a new ticket and return its id if successful.
@@ -523,8 +508,6 @@ class CartmanApp(object):
         usage: cm new [owner]
 
         """
-        self.login()
-
         template = self.resolve_template()
 
         if not template:
@@ -548,7 +531,7 @@ class CartmanApp(object):
         while not valid:
             # Get the properties at each iteration, in case an admin updated
             # the list in the mean time.
-            properties = self._get_properties()
+            properties = self.get_properties()
 
             # Assume the user will produce a valid ticket
             valid = True
@@ -562,7 +545,7 @@ class CartmanApp(object):
             fp.write(body)
 
             fp.close()
-            self._editor(filename)
+            self.editor(filename)
 
             # Use the email parser to get the headers.
             ep = email.parser.Parser()
@@ -619,7 +602,7 @@ class CartmanApp(object):
                     print(u" - {}".format(error))
 
                 try:
-                    self._input("\n-- Hit Enter to return to editor, "\
+                    self.input("\n-- Hit Enter to return to editor, "\
                               "^C to abort --\n")
                 except KeyboardInterrupt:
                     raise exceptions.FatalError("ticket creation interrupted")
@@ -652,7 +635,10 @@ class CartmanApp(object):
         r = self.post("/newticket", fields_data)
 
         if r.status_code != 200:
-            raise exceptions.RequestException("unable to create new ticket.")
+            message = text.extract_message(r.text)
+            if not message:
+                message = "unable to create new ticket"
+            raise exceptions.RequestException(message)
 
         try:
             ticket_id = int(r.url.split("/")[-1])
@@ -660,4 +646,144 @@ class CartmanApp(object):
             raise exceptions.RequestException("returned ticket_id is invalid.")
 
         self.open_in_browser_on_request(ticket_id)
-        print("ticket #{} created".format(ticket_id))
+
+        return ["ticket #{} created".format(ticket_id)]
+
+    def run_open(self, ticket_id):
+        """Open a ticket in your browser.
+
+        usage: cm open ticket_id
+        """
+        ticket_id = text.validate_id(ticket_id)
+
+        self.open_in_browser(ticket_id)
+
+    def run_properties(self):
+        """Lists the system's properties (Milestone, Component, etc.).
+
+        usage: cm properties
+
+        """
+        properties = self.get_properties()
+
+        return [
+            ui.title("Milestones"),
+            ", ".join(properties["milestone"]["options"]),
+            "",
+            ui.title("Components"),
+            ", ".join(properties["component"]["options"]),
+            "",
+            ui.title("Status"),
+            ", ".join(properties["status"]["options"]),
+            "",
+            ui.title("Priority"),
+            ", ".join(properties["priority"]["options"]),
+            "",
+        ]
+
+    def run_report(self, report_id=None):
+        """List tickets from a given report number.
+
+        usage: cm report report_id
+
+        """
+        output = []
+        report_id = text.validate_id(report_id)
+
+        query_string = "/report/{}?format=tab".format(report_id)
+
+        for t in self.get_tickets(query_string):
+            output.append(t.format_title())
+
+        return output
+
+    def run_reports(self):
+        """List reports available in the system.
+
+        usage: cm reports
+
+        """
+        output = []
+
+        for d in self.get_dicts("/report?format=tab"):
+            output.append("#{report}. {title}".format(**d))
+
+        return output
+
+    def run_search(self, *terms):
+        """Search for tickets using the given terms.
+
+        TODO: multi-page search results.
+
+        usage: cm search term
+
+        """
+        output = []
+        query_string = "/search?q={}".format("+".join(terms))
+
+        r = self.get(query_string)
+        for ticket_id, description in text.extract_search_results(r.text):
+            output.append("#{}. {}".format(ticket_id, description))
+
+        return output
+
+    def run_status(self, ticket_id, status=None):
+        """Updates the status of a ticket.
+
+        usage: cm status ticket_id [new_status]
+
+        """
+        output = []
+        ticket_id = text.validate_id(ticket_id)
+
+        # Get all the available actions for this ticket
+        r = self.get("/ticket/{}".format(ticket_id))
+        statuses = text.extract_statuses(r.text)
+
+        # Just display current status.
+        if not status:
+            status = self.extract_status_from_ticket_page(r.text)
+            output.append("Current status: {}".format(status))
+            if statuses:
+                output.append("Available statuses: {}".format(", ".join(statuses)))
+            return output
+
+        if not status:
+            raise exceptions.FatalError("bad status (acceptable: {})"
+                                        .format(", ".join(statuses)))
+
+        if self.message:
+            comment = self.message
+        elif self.add_comment:
+            comment = self._read_comment()
+        else:
+            comment = ""
+
+        # Not having a value for submit causes Trac to ignore the request.
+        data = {
+            "action": status,
+            "comment": comment,
+            "submit": "anything",
+        }
+        data.update(self._extract_timestamps(r.text))
+
+        r = self.post("/ticket/{}".format(ticket_id), data)
+
+    def run_view(self, ticket_id):
+        """Display a ticket summary.
+
+        usage: cm view ticket_id
+
+        """
+        ticket_id = text.validate_id(ticket_id)
+
+        query_string = "/ticket/{}?format=tab".format(ticket_id)
+
+        t = next(self.get_tickets(query_string))
+        title = t.format_title()
+
+        return [
+            ui.title(title),
+            "",
+            t.description,
+        ]
